// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// Copyright 2005 David le Roux

// For a description of the algorithm implemented here, see:
// http://www-cs-faculty.stanford.edu/~knuth/papers/dancing-color.ps.gz
// http://spivey.oriel.ox.ac.uk/~mike/comp2005/results.html

#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <math.h>
#include <ctime>
#include <cstdlib>
#include <time.h>
#include "SudokuSolver.h"
using namespace std;

extern long maxSolutions;

void SudokuSolver::putNodeBack(Node* node)
{
	node->columnHeader->size++;
	node->up->down = node;
	node->down->up = node;
}

void SudokuSolver::removeNode(Node* node)
{
	if(node == 0)
	cout << "trying to remove null node ???? " << endl;
	
	node->columnHeader->size--;
	node->up->down = node->down;	
	node->down->up = node->up;
}

void SudokuSolver::deleteNode(Node* node)//, Node** rowHelperArr)
{
	node->columnHeader->size--;
	
	node->up->down = node->down;	
	node->down->up = node->up;
	node->left->right = node->right;
	node->right->left = node->left;
	delete node;
}

void SudokuSolver::putRowBack(Node* node)
{
	Node* originalNode = node;
	node = node->left;
	while(node != originalNode)
	{
		putNodeBack(node);
		node = node->left;
	}	
}

void SudokuSolver::removeRow(Node* node)
{
	
	Node* originalNode = node;
	node = node->right;
	while(node != originalNode)
	{
		removeNode(node);
		node = node->right;
	}
	
}
void SudokuSolver::deleteRow(Node* node, Node** rowHelperArr)
{
/*	if(rowHelperArr[node->rowNumber] == 0)
	{
		cout << "wtf" << endl;
		exit(1);
	}*/
	
	rowHelperArr[node->rowNumber] = 0;
	
	
	Node* originalNode = node;
	node = node->right;
	
	while(node != originalNode)
	{
		if(node->columnHeader->size == 1)
		{
			impossible = true;
			return;
//			cout << "impossible to solve" << endl;//there must be at least one node in each column or the exact cover has no solution
//			exit(1);
		}
		
		Node* nextNode = node->right;
		deleteNode(node);//, rowHelperArr);
		node = nextNode;
	}
	deleteNode(originalNode);//, rowHelperArr);
	
}

void SudokuSolver::rowInMeshDefinite( int rowInMesh, Node** rowHelperArr)
{
	Node* node = rowHelperArr[rowInMesh];
	if(node == 0) //a definite row has already been deleted because of other givens
	{
		impossible = true;
		return;
		//cout << "impossible input" << endl;
		//exit(1);
	}
	
	Node* originalNode = node;

	do //now horizontally traverse this whole row
	{
		ColumnHeader* col = node->columnHeader;
    
    Node* hunting = col->down;
    while(hunting != col)
		{
      Node* huntingTmp = hunting->down;
			if(hunting != node)
			{
				deleteRow(hunting, rowHelperArr);
				if(impossible)
					return;
			}
			hunting = huntingTmp;
		}
		node = node->right;		
	} 
	while(node != originalNode);
		
}
Node* SudokuSolver::findLastNodeInColumn(int col, ColumnHeader** helperArr)
{
	Node* node = helperArr[col];
	while(node->down != 0)
		node = node->down;
	return node;
}

void SudokuSolver::createNodeRow(int row, int col1, int col2, int col3, int col4, ColumnHeader** helperArr, Node** rowHelperArr)
{
	
	Node* above = findLastNodeInColumn(col1, helperArr);
	Node* node1 = new Node(above, 0, helperArr[col1], row);
	
	above = findLastNodeInColumn(col2, helperArr);
	Node* node2 = new Node(above, node1, helperArr[col2], row);
	
	above = findLastNodeInColumn(col3, helperArr);
	Node* node3 = new Node(above, node2, helperArr[col3], row);
	
	above = findLastNodeInColumn(col4, helperArr);
	Node* node4 = new Node(above, node3, helperArr[col4], row);
	
	node4->right = node1;
	node1->left = node4;
	
	rowHelperArr[row] = node1;
}
void SudokuSolver::deleteMesh()
{
	Node* node = root->right;
	while(node != root)
	{
		Node* rightOfNode = node->right;
		Node* columnNode = node->down;
		while(columnNode != node)
		{
			Node* belowColumnNode = columnNode->down;
			delete columnNode;
			columnNode = belowColumnNode;
		}
		delete node;
		node = rightOfNode;
	}
  delete root;
  root = 0;
}

bool SudokuSolver::constructGraph(vector<int> puzzle)
{
	solutionCount = 0;
	stopSearching = false;
	
	impossible = false;
	sumOfSizes = 0;
	deterministic = true;
	rowStack.clear();
	solutions.clear();
	
	if(root != 0)
		deleteMesh();
	//cout << "mesh deleted" << endl;
	side = (int)sqrt((float)puzzle.size());
	 int sideSquared = side*side;
	 int squareRootOfSide = ( int)sqrt((float)side);
	 int columns = 4*sideSquared;
	 int rows = side * sideSquared;
		
	//build the sparse mesh (double linked up/down and left/right)
	
	ColumnHeader** helperArr = new ColumnHeader*[columns];
	Node** rowHelperArr = new Node*[rows];
	for( int i=0;i<rows;i++)
		rowHelperArr[i] = 0;	
	
	for( int i=0;i<columns;i++)
	{
		if(i==0)
		{
			root = new ColumnHeader(0);
			helperArr[0] = new ColumnHeader(side);
			root->right = helperArr[0];
			helperArr[0]->left = root;
		}
		else
			helperArr[i] = new ColumnHeader(helperArr[i-1], side);
		
		
		if(i==columns-1)
		{
			helperArr[i]->right = root;
			root->left = helperArr[i];
		}
	}

	//as an example I will quickly talk about the 9x9 sudoku puzzle:
	//each row in mesh is in this pattern: 1 in square 1, 2 in square 1, 3 in square 1 ... 9 in square 81
	//columns in mesh are the constraints, there are 4 sections:
	//section 1: 81 columns in mesh are: square 1, square 2 .. square 81 ie can only have one thing in a square
	//section 2: 81 columns in mesh are: 1 in row 1, 2 in row 1, 3 in row 1, ... 9 in row 9 ie each row in puzzle must have all
	//section 3: 81 columns in mesh are: 1 in col 1, 2 in col 1, 3 in col 1, ... 9 in col 9 ie each col in puzzle must have all
	//section 4: 81 columns in mesh are: 1 in block 1, 2 in block 1 ... 9 in block 9 ie each block in puzzle must have all 
	
	for( int i=0;i<rows;i++)
	{
		 int square = i/side; //which square (out of, say 81)
		
		 int row = i/sideSquared; //which row in the puzzle
		 int number = i - (side * square); //what number in the puzzle (say 0-8)
		 int rowPlusNumber = sideSquared + row*side + number;
		
		 int col = square - (square/side)*side; //which column in the puzzle (note integer division)
		 int colPlusNumber = 2*sideSquared + col*side + number;
		
		 int block = (row/squareRootOfSide)*squareRootOfSide + col/squareRootOfSide; //which block in the puzzle
		 int blockPlusNumber = 3*sideSquared + block*side + number;
		
		 createNodeRow(i, square, rowPlusNumber, colPlusNumber, blockPlusNumber, helperArr, rowHelperArr);
	}
	
	for(int i=0; i < columns; i++)
	{
		Node* last = findLastNodeInColumn(i, helperArr);
		last->down = helperArr[i];
		helperArr[i]->up = last;
	}

	//now if we solve this, we should get all the possible sudoku puzzles for this sudoku puzzle size
	//but because we have knowns, we will remove all unnecessary rows from the mesh
	for(int i=0;i<puzzle.size();i++)
	{
		int num = puzzle[i];
		if(num > 0)
		{
		
			//checking done in SudokuWidget
//			if(num < side+1)
//			{
				 int rowInMesh = i*side + (num-1);
				rowInMeshDefinite(rowInMesh, rowHelperArr);
			    if(impossible)
					return false;
//			}
//			else
//			{
//				cout << "Error: number " << num << " too big for this puzzle" << endl;
//				exit(1);
//			}
		}
	}

	delete[] helperArr;
	delete[] rowHelperArr;
	
	return true;
}



void SudokuSolver::putSolutionIntoVector()
{
	
		vector<SquareAndNum> solution;
		for(int i=0; i<rowStack.size();i++)
		{
			int square = rowStack[i]/side;
			int number = rowStack[i] - (side * square);
			SquareAndNum s(square, number+1);
			solution.push_back(s);	
		}
		for(int i=0; i < ((int)solution.size()-1); i++)
		{
			for(int j=i+1; j < solution.size(); j++)
			{
				if(solution[j].square < solution[i].square)
				{
					SquareAndNum s = solution[i];
					solution[i] = solution[j];
					solution[j] = s;
				}
			}
		}
		vector<int> readableSolution;
		for(int i=0; i < solution.size(); i++)
		{
//			cout << solution[i].number << " ";
//			if((i+1)%side == 0)
//				cout << endl;
			readableSolution.push_back(solution[i].number);
		}
//		cout << endl;
		solutions.push_back(readableSolution);
		solutionCount++;
		if(solutionCount == maxSolutions)
			stopSearching = true;

}

void SudokuSolver::putSolutionIntoVectorForGen()
{
		vector<SquareAndNum> solution;

		vector<int> generatedPuzzle;
		
		for(int i=(int)rowStackForGen.size()-1; i>=0 ;i--)
		{
			int square = rowStackForGen[i].rowNum/side;
			int number = rowStackForGen[i].rowNum - (side * square);
			SquareAndNum s(square, number+1);
			solution.push_back(s);	
			if(rowStackForGen[i].colSize > 1)
				generatedPuzzle.push_back(s.number);
			else
				generatedPuzzle.push_back(0);
/*			if(rowStackForGen[i].colSize > 1)
			{
				hadAGreaterThanOne = true;
			}
			if(hadAGreaterThanOne)
				generatedPuzzle.push_back(s.number);
			else
				generatedPuzzle.push_back(0);	*/
		}
		
		//now sort the solution and the generated puzzle
		for(int i=0; i < ((int)solution.size()-1); i++)
		{
			for(int j=i+1; j < solution.size(); j++)
			{
				if(solution[j].square < solution[i].square)
				{
					SquareAndNum s = solution[i];
					solution[i] = solution[j];
					solution[j] = s;
					
					int num = generatedPuzzle[i];
					generatedPuzzle[i] = generatedPuzzle[j];
					generatedPuzzle[j] = num;
				}
			}
		}
//		cout << "Here is the generated easy puzzle (deterministic):" << endl;
		generatedPuzzleField->setPuzzle(generatedPuzzle);
		
/*		for(int i=0; i < generatedPuzzle.size(); i++)
		{
			if(generatedPuzzle[i]==0)
				cout << ".";
			else
				cout << generatedPuzzle[i];
			cout << " ";
			if((i+1)%side == 0)
				cout << endl;
		}
		cout << endl;*/
//		cout << "Here is the solution to the problem:" << endl;
		vector<int> readableSolution;
		for(int i=0; i < solution.size(); i++)
		{
//			cout << solution[i].number << " ";
//			if((i+1)%side == 0)
//				cout << endl;
			readableSolution.push_back(solution[i].number);
		}
//		cout << endl;
//		solutions.push_back(readableSolution);
		generatedPuzzleField->setSolution(readableSolution);
		generatedPuzzleField->setDeterministic(true);

}

int SudokuSolver::countToTreeNodeRoot(TreeNode* treeNode)
{
	int count = 0;
	while(treeNode->parent != 0)
	{
		count++;
		treeNode = treeNode->parent;
	}
	return count;
}
void SudokuSolver::putLastSolutionIntoVector(TreeNode* treeNode)
{
	//cout << "got into putLastSolutionIntoVector" << endl;
		int countToSuitableNode = countToTreeNodeRoot(treeNode);
		
		vector<SquareAndNum> solution;
		vector<int> generatedPuzzle;
		
		for(int i=0; i < lastSolution.size(); i++)
		{
			int square = lastSolution[i].rowNum/side;
			int number = lastSolution[i].rowNum - (side * square);
			SquareAndNum s(square, number+1);
			solution.push_back(s);	
			if(lastSolution[i].colSize > 1 && i < countToSuitableNode)
				generatedPuzzle.push_back(s.number);
			else
				generatedPuzzle.push_back(0);
		}
		
		//now sort the solution and the generated puzzle
		for(int i=0; i < ((int)solution.size()-1); i++)
		{
			for(int j=i+1; j < solution.size(); j++)
			{
				if(solution[j].square < solution[i].square)
				{
					SquareAndNum s = solution[i];
					solution[i] = solution[j];
					solution[j] = s;
					
					int num = generatedPuzzle[i];
					generatedPuzzle[i] = generatedPuzzle[j];
					generatedPuzzle[j] = num;
				}
			}
		}
		generatedPuzzleField->setPuzzle(generatedPuzzle);
/*		cout << "Here is the generated hard puzzle (non-deterministic, but with a unique solution):" << endl;
		for(int i=0; i < generatedPuzzle.size(); i++)
		{
			if(generatedPuzzle[i]==0)
				cout << ".";
			else
				cout << generatedPuzzle[i];
			cout << " ";
			if((i+1)%side == 0)
				cout << endl;
		}
		cout << endl; */
		
//		cout << "Here is the solution to the problem:" << endl;
		vector<int> readableSolution;
		for(int i=0; i < solution.size(); i++)
		{
//			cout << solution[i].number << " ";
//			if((i+1)%side == 0)
//				cout << endl;
			readableSolution.push_back(solution[i].number);
		}
//		cout << endl;
		generatedPuzzleField->setSolution(readableSolution);
		generatedPuzzleField->setDeterministic(false);
}

void SudokuSolver::putSolutionIntoLastSolution()
{
	lastSolution = rowStackForGen;
}

ColumnHeader* SudokuSolver::chooseColumn()
{
	ColumnHeader* column = (ColumnHeader*)root->right;
	ColumnHeader* chosenColumn = column;
	int smallestSize = chosenColumn->size;

	column = (ColumnHeader*)column->right;
	while(column != root)
	{
		if(column->size < smallestSize)
		{
			smallestSize = column->size;
			chosenColumn = column;
		}
		column = (ColumnHeader*)column->right;
	}
	return chosenColumn;
}

ColumnHeader* SudokuSolver::chooseColumnRandomly()
{
	ColumnHeader* column = (ColumnHeader*)root->right;
	
	int smallestSize = column->size;
	column = (ColumnHeader*)column->right;
	
	while(column != root)
	{
		if(column->size < smallestSize)
			smallestSize = column->size;
		column = (ColumnHeader*)column->right;
	}
	
	vector<ColumnHeader*> possibilities;
	column = (ColumnHeader*)column->right;
	
	while(column != root)
	{
		if(column->size == smallestSize)
			possibilities.push_back(column);
		column = (ColumnHeader*)column->right;
	}
	int chosen = rand()%possibilities.size();
	return possibilities[chosen];
}
void SudokuSolver::unCoverColumn(ColumnHeader* column)
{
	column->right->left = column;
	column->left->right = column;
	Node* row = column->up;
	while(row != column)
	{
		putRowBack(row);
		row = row->up;
	}
}
void SudokuSolver::coverColumn(ColumnHeader* column)
{
	
	column->left->right = column->right;
	column->right->left = column->left;
	Node* row = column->down;
	while(row != column)
	{
		removeRow(row);
		row = row->down;
	}
}

void SudokuSolver::search()
{
	if(stopSearching)
		return;
	
	if(root->right == root)
	{
		putSolutionIntoVector();
		return;
	}
	
	ColumnHeader* column = chooseColumn();
	if(column->size == 0)
	{	
		return;
	}
	if(column->size > 1)
		deterministic = false;
		
	sumOfSizes += column->size;
		
	coverColumn(column);
	
	Node* row = column->down;
	while(row != column)
	{
		rowStack.push_back(row->rowNumber);
		
		Node* originalNode = row;
		Node* node = originalNode->right;
		
		while(node != originalNode)
		{
			coverColumn(node->columnHeader);
			node = node->right;
		}
		
		search();
		
		rowStack.pop_back();
		
		node = node->left;
		while(node != originalNode)
		{
			unCoverColumn(node->columnHeader);
			node = node->left;
		}
		
		row = row->down;
	}
	
	unCoverColumn(column);
}

void SudokuSolver::searchEasyPuzzle()
{
	if(foundSolution)
		return;
		
	if(root->right == root)
	{
		putSolutionIntoVectorForGen();
		foundSolution = true;
		return;
	}

	ColumnHeader* column = chooseColumnRandomly();

	if(column->size == 0)
	{
		return;
	}
		
	coverColumn(column);

	Node* row = column->down;
	while(row != column)
	{
		RowNumAndColSize tmp(row->rowNumber,column->size);
		rowStackForGen.push_back(tmp);

		Node* originalNode = row;
		Node* node = originalNode->right;

		while(node != originalNode)
		{
			coverColumn(node->columnHeader);
			node = node->right;
		}

		searchEasyPuzzle();

		rowStackForGen.pop_back();

		node = node->left;
		while(node != originalNode)
		{
			unCoverColumn(node->columnHeader);
			node = node->left;
		}

		row = row->down;
	}

	unCoverColumn(column);
}
void SudokuSolver::bubbleSolutionDownTree(TreeNode* treeNode)
{
	while(treeNode->parent != 0)
	{
		treeNode = treeNode->parent;
		if(treeNode->numSolutions < 2)
			treeNode->numSolutions++;
	}
}

void SudokuSolver::checkNodeForHardPuzzle(TreeNode* treeNode)
{
	if(treeNode->numBranches > 1 && treeNode->numBranches == treeNode->numExplored && treeNode->numSolutions ==1 && sizeCount > minimumSizeCount)
	{
		foundHardPuzzle = true;
		putLastSolutionIntoVector(treeNode);
	}
}
void SudokuSolver::bubbleExploredDownTree(TreeNode* treeNode)
{
	while(treeNode->parent != 0)
	{
		treeNode = treeNode->parent;
		treeNode->numExplored++;
		checkNodeForHardPuzzle(treeNode);
		if(foundHardPuzzle)
			break; 
		if(treeNode->numBranches > treeNode->numExplored)
			break;
	}	
}

void SudokuSolver::foundASolution(TreeNode* treeNode)
{
	bubbleSolutionDownTree(treeNode);
	bubbleExploredDownTree(treeNode);
}

void SudokuSolver::searchMediumPuzzle(TreeNode* treeNode)
{	
	if(foundHardPuzzle)
	{
		return;
	}
			
	if(root->right == root)
	{
		putSolutionIntoLastSolution();
		foundASolution(treeNode);
		return;
	}

	ColumnHeader* column = chooseColumnRandomly();

	if(column->size == 0)
	{
		bubbleExploredDownTree(treeNode);
		return;
	}
	else
		sizeCount += column->size;
	
	treeNode->numBranches = column->size;	
	coverColumn(column);

	Node* row = column->down;
	while(row != column)
	{
		TreeNode* treeNodeChild = new TreeNode(treeNode);
		treeNodesToDelete.push_back(treeNodeChild);
		
		RowNumAndColSize tmp(row->rowNumber,column->size);
		rowStackForGen.push_back(tmp);

		Node* originalNode = row;
		Node* node = originalNode->right;

		while(node != originalNode)
		{
			coverColumn(node->columnHeader);
			node = node->right;
		}

		searchMediumPuzzle(treeNodeChild);

		rowStackForGen.pop_back();

		node = node->left;
		while(node != originalNode)
		{
			unCoverColumn(node->columnHeader);
			node = node->left;
		}

		row = row->down;
	}

	unCoverColumn(column);
}

GeneratedPuzzle*  SudokuSolver::generateEasyPuzzle(int squareRootOfSide)
{
	generatedPuzzleField->clear();
	
    rowStackForGen.clear();
    foundSolution = false;
  
    //construct an empty puzzle
    vector<int> puzzle;
    int side = squareRootOfSide*squareRootOfSide;
    for(int i=0; i<side*side; i++)
  	puzzle.push_back(0);

    constructGraph(puzzle);
  	searchEasyPuzzle();
//	cout << "about to return generatedPuzzleField" << endl;
	return generatedPuzzleField;
}

GeneratedPuzzle*  SudokuSolver::generateMediumPuzzle(int squareRootOfSide)
{
	sizeCount = 0;
	minimumSizeCount = 0;
	generatedPuzzleField->clear();
	treeNodesToDelete.clear();
	
    rowStackForGen.clear();
    foundHardPuzzle = false;
  
    //construct an empty puzzle
    vector<int> puzzle;
    int side = squareRootOfSide*squareRootOfSide;
    for(int i=0; i<side*side; i++)
  	  puzzle.push_back(0);

    constructGraph(puzzle); 
    TreeNode* root = new TreeNode();
    treeNodesToDelete.push_back(root);
    searchMediumPuzzle(root);
    //cout << "tree nodes to delete: " << treeNodesToDelete.size();
    for(int i=0; i < treeNodesToDelete.size(); i++)
  	  delete treeNodesToDelete[i];
	
	return generatedPuzzleField;
}
GeneratedPuzzle*  SudokuSolver::generateHardPuzzle(int squareRootOfSide)
{
	sizeCount = 0;
	int sideSquared = squareRootOfSide*squareRootOfSide*squareRootOfSide*squareRootOfSide;
	minimumSizeCount = (int)(0.25*sideSquared) + sideSquared;
	generatedPuzzleField->clear();
	treeNodesToDelete.clear();
	
    rowStackForGen.clear();
    foundHardPuzzle = false;
  
    //construct an empty puzzle
    vector<int> puzzle;
    int side = squareRootOfSide*squareRootOfSide;
    for(int i=0; i<side*side; i++)
  	  puzzle.push_back(0);

    constructGraph(puzzle); 
    TreeNode* root = new TreeNode();
    treeNodesToDelete.push_back(root);
    searchMediumPuzzle(root);
    //cout << "tree nodes to delete: " << treeNodesToDelete.size();
    for(int i=0; i < treeNodesToDelete.size(); i++)
  	  delete treeNodesToDelete[i];
	
	return generatedPuzzleField;
}

SudokuSolver::SudokuSolver()
{
	root = 0;
	generatedPuzzleField = new GeneratedPuzzle();
}
SolutionStruct* SudokuSolver::solutionsForPuzzle(vector<int> puzzle)
{
		if(constructGraph(puzzle)) //this sets the root
			search(); //this sets solutions	
		return new SolutionStruct(solutions, deterministic,impossible);				
}
