// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// Copyright 2005 David le Roux

#ifndef _SUDOKUSOLVER_H_
#define _SUDOKUSOLVER_H_
#include <vector>
using namespace std;

class ColumnHeader;

class Node
{
 public:
	Node* left;
	Node* right;
	Node* up;
	Node* down;
	ColumnHeader* columnHeader;
	int rowNumber;

	Node()
	{
	}
	Node(Node* above, Node* toTheLeft, ColumnHeader* header,  int rowNum)
	{
		up = above;
		up->down = this;
		left = toTheLeft;
		if(left!=0)
			left->right = this;
		right = 0;
		down = 0;
		
		columnHeader = header;
//		columnHeader->size++; //would be nice to do this rather than construct a column header with the size
//		but I don't feel like fixing the forward ref
		rowNumber = rowNum;
	}
	
};

class ColumnHeader : public Node
{
 public:
	 int size;
	
	ColumnHeader(int s)
	{
		size = s;
		left = 0;
		right = 0;
		up = 0;
		down =0;
		columnHeader = 0; //not used
		rowNumber = 0; //not used
	}
	
	ColumnHeader(ColumnHeader* prev, int s)
	{
		size = s;
		up = 0;
		down = 0;
		right = 0;
		prev->right = this;
		left = prev;
		columnHeader = 0; //not used
		rowNumber = 0; //not used
	}
};

class TreeNode
{
 public:
	TreeNode* parent;
	short numBranches;
	short numExplored;
	short numSolutions;
	TreeNode()
	{
		parent = 0;
		numBranches = 0;
		numExplored = 0;
		numSolutions = 0;
	}
	TreeNode(TreeNode* p)
	{
		parent = p;
		numBranches = 0;
		numExplored = 0;
		numSolutions = 0;
	}
	
};

struct RowNumAndColSize
{
	int rowNum;
	int colSize;
	RowNumAndColSize(int r, int c)
	{
		rowNum = r;
		colSize = c;
	}
};

struct SquareAndNum
{
	int square;
	int number;
	SquareAndNum(int s, int n)
	{
		square = s;
		number = n;
	}
};

struct SolutionStruct
{
	vector< vector<int> > solutions;
	bool deterministic;
	bool impossible;
	SolutionStruct(vector< vector<int> > s, bool d, bool i)
	{
		solutions = s;
		deterministic = d;
		impossible = i;
	}
	SolutionStruct()
	{
	}
/*	void setParameters(SolutionStruct* other)
	{
		solutions = other->solutions;
		deterministic = other->deterministic;
		impossible = other->impossible;
	}*/
/*	void setParameters(vector< vector<int> > s, bool d, bool i)
	{
		solutions = s;
		deterministic = d;
		impossible = i;
	}*/
};
struct GeneratedPuzzle
{
	vector<int> puzzle;
	vector<int> solution;
	bool deterministic;
	
	GeneratedPuzzle()
	{
		puzzle.clear();
		solution.clear();
		deterministic = false;
	}
	void setPuzzle(vector<int> p)
	{
		puzzle = p;
	}
	void setSolution(vector<int> s)
	{
		solution = s;
	}
	void setDeterministic(bool d)
	{
		deterministic = d;
	}
	void clear()
	{
		puzzle.clear();
		solution.clear();
		deterministic = false;
	}
};
class SudokuSolver
{
	public:
		SudokuSolver();
		SolutionStruct* solutionsForPuzzle(vector<int> puzzle);
		GeneratedPuzzle* generateEasyPuzzle(int squareRootOfSide);
		GeneratedPuzzle*  generateMediumPuzzle(int squareRootOfSide);
		GeneratedPuzzle*  generateHardPuzzle(int squareRootOfSide);
	private:
		ColumnHeader* root;
		vector<int> rowStack;
		vector<RowNumAndColSize> rowStackForGen;
		vector<RowNumAndColSize> lastSolution;
		int side;
		vector< vector<int> > solutions;
		bool foundSolution;
		bool deterministic;
		long solutionCount;
		bool foundHardPuzzle;
		vector<TreeNode*> treeNodesToDelete;
		long sumOfSizes;
		bool impossible;
		
		void putNodeBack(Node* node);
		void removeNode(Node* node);
		void deleteNode(Node* node);//, Node** rowHelperArr);
		void putRowBack(Node* node);
		void removeRow(Node* node);
		void deleteRow(Node* node, Node** rowHelperArr);
		void rowInMeshDefinite( int rowInMesh, Node** rowHelperArr);
		Node* findLastNodeInColumn(int col, ColumnHeader** helperArr);
		void createNodeRow(int row, int col1, int col2, int col3, int col4, ColumnHeader** helperArr, Node** rowHelperArr);
		void deleteMesh();
		bool constructGraph(vector<int> puzzle);
		void putSolutionIntoVector();
		void putSolutionIntoVectorForGen();
		int countToTreeNodeRoot(TreeNode* treeNode);
		void putLastSolutionIntoVector(TreeNode* treeNode);
		void putSolutionIntoLastSolution();
		ColumnHeader* chooseColumn();
		ColumnHeader* chooseColumnRandomly();
		void unCoverColumn(ColumnHeader* column);
		void coverColumn(ColumnHeader* column);
		void search();
		void searchEasyPuzzle();
		void bubbleSolutionDownTree(TreeNode* treeNode);
		void checkNodeForHardPuzzle(TreeNode* treeNode);
		void bubbleExploredDownTree(TreeNode* treeNode);
		void foundASolution(TreeNode* treeNode);
		void searchMediumPuzzle(TreeNode* treeNode);
//		void solvePuzzle();

		bool stopSearching;
		GeneratedPuzzle* generatedPuzzleField;
		int sizeCount;
		int minimumSizeCount;
	

};
#endif /*_SUDOKUSOLVER_H_*/
