// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Library General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
// Copyright 2005 David le Roux

#include "SudokuWidget.h"
#include "SudokuSolver.h"
#ifdef QT3
 #include <qfiledialog.h>
#else
 #include <QtGui>
 #include <QFileDialog>
 #include <QTextStream>
 #include <QRegExp>
 #include <QImage>
#endif

#include <math.h>
#include <iostream>
using namespace std;

long maxSolutions = 1001;

SudokuWidget::SudokuWidget(QAction* u,QAction* r,QStatusBar* s)
    : QWidget(0)
{
    undoAct = u;
    redoAct = r;
    undoAct->setEnabled(false);
    redoAct->setEnabled(false);
	statusBar = s;

//    lastKey = -1;
    setMouseTracking(true);
    QPalette palette;
    palette.setColor(backgroundRole(), QColor(Qt::white));
    setPalette(palette);
    
    displayFont.setFamily("Arial");
 //   displayFont.setPixelSize(10);

    highlight = new Highlight();
//    setAttribute(Qt::WA_PaintOnScreen,true);
//    setAttribute(Qt::WA_PaintOutsidePaintEvent);  
    
    side = 0;
    setSide(9);
	
	sudokuSolver = new SudokuSolver();
	lastLocation = "";
	lastSavedLocation = "";
	solutionStruct = 0;
	
	solutionDialog = new SolutionDialog(this);
	
	browser = new QTextBrowser();
	browser->setHtml("<b>Sudoku Swiss Army Knife</b>" \
	"<p>" \
	"The aim of Sudoku is to get one of each digit in each row column and block." \
	"</p>" \
	"<p>" \
	"Press the left mouse button on a number to 'pencil' it in." \
	"</p>" \
	"<p>" \
	"Press the right mouse button once you are happy with your choice." \
	"</p>" \
	"<p>" \
	"You can not only play generated games, but you can enter them yourself, " \
	"by right clicking the relevant digits and then clicking on 'Game', 'Fix'." \
	"</p>" \
	);
}
void SudokuWidget::setSide(int s, bool allowUndo)
{
	vector<RowCol> toSave;
	
  gameFixed = false;
  int root = sqrt((float)s);
  if(root*root == s)
  {
    for(int i=0; i<side; i++)
    {
      for(int j=0; j<side; j++)
      {
		  if(allowUndo)
			  backupSquare(i,j);
		  
            grid[i][j]->numbers.clear();
            delete grid[i][j];
      }  
      delete[] grid[i];
    } 
    if(side > 0)
        delete[] grid;
      
	
	int oldSideSquared = side*side;
    side = s;
    squareRootOfSide = root;
	sideSquared = side*side;
    int minimumSquareSize;
//	int maximumSquareSize = 150;
    if(side < 10)
        minimumSquareSize = 30;
    else
        minimumSquareSize = 60;
    setMinimumSize(side*minimumSquareSize,side*minimumSquareSize);
	
	//setMaximumSize(side*maximumSquareSize,side*maximumSquareSize);
    
    grid = new Square**[side];
    for(int i=0; i<side; i++)
    {
      grid[i] = new Square*[side];
      for(int j=0; j<side; j++)
      {
        grid[i][j] = new Square();
        for(int k=0; k<side; k++)
          grid[i][j]->numbers.push_back(off); 
		
		if(allowUndo)
			toSave.push_back(RowCol(i,j));
      }     
    }  
    numbersAsStrings.clear();
    for(int i=1; i <= side; i++)
        numbersAsStrings.push_back(QString::number(i));
    
	if(allowUndo)
		saveChangedSquares(toSave);
  
	if(!allowUndo)
	{
		 //cout << "about to delte undoStack" << endl;
		  undoPosition = -1;
		  //now clear out old undoStack
		  for(int i=0; i<undoStack.size();i++)
		  {
			  for(int j=0; j<oldSideSquared; j++)
			  {
				  if(undoStack[i]->squares[j] != 0)
					  delete undoStack[i]->squares[j];
			  }
			  delete undoStack[i];
		  }
		  //cout << "deleted undostack" << endl;
		  undoStack.clear();
		  //put an empty starting grid onto the undoStack
		  SquareArray* squareArray = new SquareArray(sideSquared);
		  undoStack.push_back(squareArray);
		  undoAct->setEnabled(false);
		  redoAct->setEnabled(false);
	}
//	else
	if(width()>side*minimumSquareSize || height()>side*minimumSquareSize)
		resize(side*minimumSquareSize,side*minimumSquareSize);
	else
		update();
  }
}  

void SudokuWidget::threeEasy()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateEasyPuzzle(3);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::threeMed()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateMediumPuzzle(3);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::threeHard()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateHardPuzzle(3);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::fourEasy()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateEasyPuzzle(4);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::fourMed()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateMediumPuzzle(4);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::fourHard()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateHardPuzzle(4);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::fiveEasy()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateEasyPuzzle(5);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::fiveMed()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateMediumPuzzle(5);
	displayGenerated(generatedPuzzle);
}
void SudokuWidget::fiveHard()
{
	GeneratedPuzzle* generatedPuzzle = sudokuSolver->generateHardPuzzle(5);
	displayGenerated(generatedPuzzle);
}

void SudokuWidget::displayGenerated(GeneratedPuzzle* g)
{
	vector<int> puzzle = g->puzzle;
//	cout << "puzzle " << endl;
//	for(int i=0; i<puzzle.size(); i++)
//		cout << puzzle[i] << endl;
//	cout << "setting side " << endl;
	setSide((int)sqrt((float)puzzle.size()),false);
	
	for(int i=0; i<puzzle.size(); i++)
	{
		if(puzzle[i] != 0)
		{
			int row = i/side;
			int col = i - row*side;
			int num = puzzle[i]-1;
			fix(row,col,num,false);
		}
	}
	actualFix(false,true);
	if(solutionStruct != 0)
			delete solutionStruct;
	vector< vector<int> > solutions; //even though there is only one
	solutions.push_back(g->solution);
//	cout << "solution " << endl;
//	for(int i=0; i< g->solution.size(); i++)
//		cout << g->solution[i] << endl;
	
	solutionStruct = new SolutionStruct(solutions,g->deterministic,false);
	statusBarMessage(solutionStruct);
}

void SudokuWidget::displayPuzzle(vector<int> puzzle)
{
	setSide((int)sqrt((float)puzzle.size()),false);
	for(int i=0; i<puzzle.size(); i++)
	{
		if(puzzle[i] != 0)
		{
			int row = i/side;
			int col = i - row*side;
			int num = puzzle[i]-1;
			fix(row,col,num,false);
		}
	}
	actualFix(false,true);
}
void SudokuWidget::displayProperFile(QString contents)
{
	if(contents.length()>0)
	{
		QStringList stringList = contents.split(QRegExp("\\n"));
		int side = (int)sqrt((float)((stringList.size()-2)/3));
		
		int squareRootOfSide = (int)sqrt((float)side);
		if(squareRootOfSide*squareRootOfSide*squareRootOfSide*squareRootOfSide != side*side)
			return;
		setSide(side,false);
		int count = 0;
		for (int i = 1; i < stringList.size()-1; i += 3)
		{
	//		cout << "at count" << count << endl;
			int row = count/side;
			int col = count - row*side;
			
			QString token = stringList.at(i);
			if(token == "true") //really fixed
			{
				int num = stringList.at(i+1).toInt();
			//	cout  << num << endl;
				fix(row,col,num-1,false);
				grid[row][col]->reallyFixed = true;
			}
			else
			{
				if(stringList.at(i+1).length() > 0)
				{
					int num = stringList.at(i+1).toInt();
					fix(row,col,num-1,false);
				}
				else
				{
					QStringList numbersOn = stringList.at(i+2).split(QRegExp(" "),QString::SkipEmptyParts);
					if(numbersOn.size() <= side)
						for(int j=0; j<numbersOn.size(); j++)
						{
							int num = numbersOn.at(j).toInt();
							
							if(num<side)
							if(grid[row][col]->numbers[num] != disabled)
							{
								grid[row][col]->numbers[num] = on;
							}
						}
				}
			}
			count++;
		}
	}
}
vector<int> SudokuWidget::readInFile(QString fileName,bool& properFileLayout,QString& contents)
{
	vector<int> puzzle;
	
	QFile file(fileName);
	if(file.open(QFile::ReadOnly))
	{
		
		QTextStream stream(&file);
		QString firstLine = stream.readLine();
		if(firstLine == "SudokuSAK")
		{	
			stream.seek(0);
			contents = stream.readAll();
			stream.seek(0);
			stream.readLine();
			
			properFileLayout = true;
			QString line = stream.readLine();
			while(!(line.isNull()))
			{
				if(line=="true")
				{
					line = stream.readLine();
					bool ok;
					int num = line.toInt(&ok);
					if(!ok)
					{
						QMessageBox::warning(this, "SudokuSAK", tr("Token: [")+line + tr("] could not be converted to an integer."));
						puzzle.clear();
						return puzzle;
					}
					else
					{
						if(num < 0)
						{
							QMessageBox::warning(this, "SudokuSAK",tr("File containted a negative number."));
							puzzle.clear();
							return puzzle;
						}
						else
						{
							puzzle.push_back(num);
						}
					}
					stream.readLine();
				}
				else
				{
					stream.readLine();
					stream.readLine();
					puzzle.push_back(0);
				}
				line = stream.readLine();
			}
			file.close();
			return puzzle;
		}
		else
		{	
			QString fileString = firstLine + stream.readAll();
			file.close();
			
			QStringList stringList = fileString.split(QRegExp("\\s+"), QString::SkipEmptyParts);
			for (int i = 0; i < stringList.size(); ++i)
			{
				int num = 0;
				QString token = stringList.at(i);
				if(token != ".")
				{
					bool ok;
					num = token.toInt(&ok);
					if(!ok)
					{
						QMessageBox::warning(this, "SudokuSAK", tr("Token: [")+token + tr("] could not be converted to an integer."));
						puzzle.clear();
						return puzzle;
					}
					else
					{
						if(num < 0)
						{
							QMessageBox::warning(this, "SudokuSAK",tr("File containted a negative number."));
							puzzle.clear();
							return puzzle;
						}
						else
							puzzle.push_back(num);
					}
				}
				else
					puzzle.push_back(num);
			}
			return puzzle;
	    }
	}
	else
	{
		QMessageBox::warning(this, "SudokuSAK", "Could not open the selected file");
		return puzzle; //empty
	}
}

void SudokuWidget::writeOutSudokuFile(QString location)
{
	QFile file(location);
	if(file.open(QFile::WriteOnly))
	{
		QTextStream stream(&file);
		stream << "SudokuSAK" << endl;
		for(int i=0; i<side; i++)
			for(int j=0; j<side; j++)
			{
				if(grid[i][j]->reallyFixed)
					stream << "true" << endl;
				else
					stream << endl;
				if(grid[i][j]->fix > 0)
					stream << grid[i][j]->fix << endl;
				else
					stream << endl;
				for(int x=0; x< grid[i][j]->numbers.size(); x++)
					if(grid[i][j]->numbers[x] == on)
						stream << x << " ";
				stream << endl;
			}				
		file.close();
	}
}
void SudokuWidget::statusBarMessage(SolutionStruct* s)
{
	QString message = "Number of solutions: ";
	if(solutionStruct->solutions.size() >= maxSolutions)
	{
		QString maxSolutionsString;
		maxSolutionsString.setNum(maxSolutions-1);
		message += "over " + maxSolutionsString;
	}
	else
	{
		QString solutionSizeString;
		solutionSizeString.setNum(solutionStruct->solutions.size());
		message += solutionSizeString;
	}
	if(solutionStruct->solutions.size() == 1)
	{
		message += "     Exact cover was: ";
		if(solutionStruct->deterministic)
			message += "deterministic";
		else
			message += "non-deterministic";
	}
	statusBar->showMessage(message);
}
void SudokuWidget::openSlot()
{
	QString s = QFileDialog::getOpenFileName(
                    this,
                    "Choose a file",
                    lastLocation,
                    "Sudoku Game Files (*.sud)");
	
	if(!s.isNull())
	{
		lastLocation = s;
		bool properFileLayout = false;
		QString contents;
		vector<int> puzzle = readInFile(s,properFileLayout,contents);
		if(puzzle.size()>0)
		{
			int side = (int)sqrt((float)puzzle.size());
			if(side > 0)
			{
				int squareRootOfSide = ( int)sqrt((float)side);
				if((squareRootOfSide*squareRootOfSide*squareRootOfSide*squareRootOfSide) == puzzle.size())
				{
					bool fine = true;
					for(int i=0; i< puzzle.size(); i++)
					{
						if(puzzle[i] > side)
						{
							fine = false;
							break;
						}
					}
					if(fine)
					{
						if(solutionStruct != 0)
							delete solutionStruct;
						solutionStruct = sudokuSolver->solutionsForPuzzle(puzzle);
						if(!(solutionStruct->impossible))
						{
							statusBarMessage(solutionStruct);
							
							if(properFileLayout)
								displayProperFile(contents);
							else
								displayPuzzle(puzzle);
						}
						else
							QMessageBox::warning(this, "SudokuSAK", "Impossible to solve (invalid input).");
					}	
					else
						QMessageBox::warning(this, "SudokuSAK", "Not a valid file");
				}
				else
					QMessageBox::warning(this, "SudokuSAK", "Not a valid file");
			}
			else
				QMessageBox::warning(this, "SudokuSAK", "Not a valid file");
		}
	}
}

void SudokuWidget::saveSlot()
{
	QString s = QFileDialog::getSaveFileName(
                    this,
                    "Choose a filename to save under",
                    lastSavedLocation,
                    "SudokuSAK (*.sud)");
	if(!s.isNull())
	{
		lastSavedLocation = s;
		
		writeOutSudokuFile(lastSavedLocation);
	}
}

void SudokuWidget::saveAsImageSlot()
{
	highlight->active = false;
	QPixmap* pixmap = new QPixmap(width(),height());
	pixmap->fill();
	QImage image = pixmap->toImage();
	paintOnDevice(&image,QRect(0,0,width(),height()));
	QString s = QFileDialog::getSaveFileName(
                    this,
                    "Choose a filename to save under",
                    lastSavedLocation,
                    "Images (*.png)");
	if(!s.isNull())
	{
		lastSavedLocation = s;
		image.save(lastSavedLocation,"PNG");
	}
	delete pixmap;
}
vector<int> SudokuWidget::puzzleFromCurrent()
{
	vector<int> puzzle;
	for(int i=0;i<side;i++)
		for(int j=0; j<side; j++)
		{
			if(grid[i][j]->reallyFixed)
				puzzle.push_back(grid[i][j]->fix);
			else
				puzzle.push_back(0);
		}
	return puzzle;
}
void SudokuWidget::solveCurrent()
{
	vector<int> puzzle = puzzleFromCurrent();
		
	if(solutionStruct != 0)
			delete solutionStruct;
	solutionStruct = sudokuSolver->solutionsForPuzzle(puzzle);
	if(!(solutionStruct->impossible))
	{
		statusBarMessage(solutionStruct);
	}
	else
		QMessageBox::warning(this, "SudokuSAK", "Impossible to solve (invalid input).");

			
}
void SudokuWidget::actualFix(bool allowUndo,bool bypassFindSolution)
{
	gameFixed = true;
  vector<RowCol> v;
	
  for(int i=0; i<side; i++)
    for(int j=0; j<side; j++)
    {
        if(grid[i][j]->fix > 0)
		{
			if(allowUndo)
				backupSquare(i,j);
            grid[i][j]->reallyFixed = true;
			if(allowUndo)
				v.push_back(RowCol(i,j));
			update(squareToPaint(i,j));
		}
    }
	if(allowUndo)
  		saveChangedSquares(v);	
	
	if(!bypassFindSolution)
		solveCurrent();
}
void SudokuWidget::fixSlot()
{
 	actualFix(true,false);
}
//void SudokuWidget::generateSlot()
//{
//}
void SudokuWidget::clearAllSlot()
{
	setSide(side,true);
}
void SudokuWidget::clearNonFixedSlot()
{
	vector<RowCol> toSave;
	
    for(int i=0; i<side; i++)
      for(int j=0; j<side; j++)
      {
		  if(!(grid[i][j]->reallyFixed))
		  {
			  backupSquare(i,j);		  
              grid[i][j]->numbers.clear();
              delete grid[i][j];
			  
			  grid[i][j] = new Square();
        	  for(int k=0; k<side; k++)
              	grid[i][j]->numbers.push_back(off); 
			  
			  toSave.push_back(RowCol(i,j));
		  }
      }  
     
	for(int i=0;i<side;i++)
		for(int j=0;j<side;j++)
			if(grid[i][j]->reallyFixed)
				fix(i, j, grid[i][j]->fix-1, false);				
	
	saveChangedSquares(toSave);
  
	update();
  
}
void SudokuWidget::digitCountSlot()
{
}
void SudokuWidget::showSolutionsSlot()
{
	if(solutionStruct != 0 && solutionStruct->solutions.size()>0)
	{
		solutionDialog->setSolutionStruct(solutionStruct);
		solutionDialog->show();
	}
	else
	{
		QMessageBox::warning(this, "SudokuSAK", tr("No solutions to show."));
	}
}

void SudokuWidget::changeBoardFromStack(int position)
{
	SquareArray* squareArray = undoStack[position];
	for(int i=0; i<sideSquared; i++)
	{
		if(squareArray->squares[i] != 0)
		{
			int row = i/side;
			int col = i - row*side;
			delete grid[row][col];
			Square* copyOfSquare = new Square(squareArray->squares[i]);
      		grid[row][col] = copyOfSquare;
      		update(squareToPaint(row,col));
		}
	}
}
void SudokuWidget::undoSlot()
{
	redoAct->setEnabled(true);
	
	changeBoardFromStack(undoPosition);
 
    undoPosition--;  
    if(undoPosition == -1)
        undoAct->setEnabled(false);
    
}
void SudokuWidget::redoSlot()
{
	undoAct->setEnabled(true);
	
	undoPosition++;
	changeBoardFromStack(undoPosition+1);
	if(undoPosition == undoStack.size()-2)
		redoAct->setEnabled(false);
}
void SudokuWidget::settingsSlot()
{
}

void SudokuWidget::manualSlot()
{
	browser->show();
}
void SudokuWidget::aboutSlot()
{
  QMessageBox::about (this, tr("About SudokuSAK"), tr("Sudoku Swiss Army Knife \n\nby David le Roux\nLicensed under the GPL"));
}

void SudokuWidget::leaveEvent ( QEvent * event )
{
 // highlight->active = false;
 // update();
}  
QRect SudokuWidget::squareToPaint(QPoint pos)
{
  int paintx = (pos.x()/squareWidth)*squareWidth+1;
  int painty = (pos.y()/squareHeight)*squareHeight+1;
  return QRect(paintx,painty,squareWidth-2,squareHeight-2);
}  

QRect SudokuWidget::squareToPaint(int row,int col)
{
  int paintx = col * squareWidth + 1;
  int painty = row * squareHeight + 1;
  return QRect(paintx,painty,squareWidth-2,squareHeight-2);
}  

void SudokuWidget::mouseMoveEvent(QMouseEvent *event)
{
    QPoint pos = event->pos();
     
    if(highlight->active)
    {
      if(!highlight->rect.contains(pos))
      {
        highlight->active = false;
        update(squareToPaint(highlight->row,highlight->col));
      }  
    }  
    
    int row = pos.y()/squareHeight;
    int col = pos.x()/squareWidth;
    if(pos.x()>=maxWidth || pos.y()>=maxHeight)
        return;
 
    Square* square = grid[row][col];
    if((square->fix > 0))
        return;
    
      int cellRow = (pos.y()-row*squareHeight)/cellHeight;
      if(cellRow >= squareRootOfSide)
            cellRow = squareRootOfSide -1;
      int cellCol = (pos.x()-col*squareWidth)/cellWidth;
      if(cellCol >= squareRootOfSide)
            cellCol = squareRootOfSide -1;
      int cellNum = squareRootOfSide*cellRow + cellCol;
      if(square->numbers[cellNum] != disabled)
      {     
        int x1 = col*squareWidth + cellCol*cellWidth;
        int y1 = row*squareHeight + cellRow*cellHeight;
        int x2 = x1 + cellWidth;
        int y2 = y1 + cellHeight;
        highlight->set(row,col,cellRow,cellCol,x1,y1,x2,y2);
        update(squareToPaint(pos)); 
      }  
      
}

void SudokuWidget::wipeOutRedos()
{
  redoAct->setEnabled(false);
//  if( undoPosition != -1)
//  {
    for(int i=undoPosition+2; i<undoStack.size(); i++)
    {
      for(int j=0; j<sideSquared; j++)
	  {
		  if(undoStack[i]->squares[j] != 0)
		  	delete undoStack[i]->squares[j];
	  }
      delete undoStack[i];
    }
    int count = undoStack.size() - undoPosition - 2;
    for(int i=0; i<count; i++)
    {
      undoStack.pop_back();
    }    
 // } 
}
void SudokuWidget::backupSquare(int row,int col)
{
	if(undoStack[undoPosition+1]->squares[row*side + col] == 0)
	{
		Square* copyOfSquare = new Square(grid[row][col]);
		undoStack[undoPosition+1]->squares[row*side + col] = copyOfSquare;
	}
}
void SudokuWidget::saveChangedSquares(vector<RowCol> v)
{
	if(v.size() == 0)
		return;
	
	wipeOutRedos();
    undoPosition++;
    undoAct->setEnabled(true);
	
	SquareArray* squareArray = new SquareArray(sideSquared);
	for(int i=0; i<v.size(); i++)
	{
		int pos = v[i].row*side + v[i].col;
		if(squareArray->squares[pos] == 0)
		{
			Square* copyOfSquare = new Square(grid[v[i].row][v[i].col]);
			squareArray->squares[pos] = copyOfSquare;
		}
	}
	undoStack.push_back(squareArray);
}
/*
void SudokuWidget::putNewSquareOnUndoStack(int row, int col)
{
	SquareArray* squareArray = new SquareArray(sideSquared);
	int pos = row*side + col;
	Square* copyOfSquare = new Square(grid[row][col]);
	squareArray->squares[pos] = copyOfSquare;
    undoStack.push_back(squareArray);
}*/
void SudokuWidget::mousePressEvent(QMouseEvent *event)
{
  QPoint pos = event->pos();
  if(pos.x()>=maxWidth || pos.y()>=maxHeight)
      return;
  int row = pos.y()/squareHeight;
  int col = pos.x()/squareWidth;
  Square* square = grid[row][col];
  if((square->fix > 0))
    return;
  
    int cellRow = (pos.y()-row*squareHeight)/cellHeight;
    if(cellRow >= squareRootOfSide)
          cellRow = squareRootOfSide -1;
    int cellCol = (pos.x()-col*squareWidth)/cellWidth;
    if(cellCol >= squareRootOfSide)
          cellCol = squareRootOfSide -1;
    int cellNum = squareRootOfSide*cellRow + cellCol;
        
    if (event->button() == Qt::LeftButton) 
    {
      if(square->numbers[cellNum] != disabled)
      {
        
		backupSquare(row,col);
		  
        if(square->numbers[cellNum] == on)
          square->numbers[cellNum] = off;
        else
          square->numbers[cellNum] = on;
		
	    vector<RowCol> v;
		v.push_back(RowCol(row,col));
        saveChangedSquares(v);
		
        update(squareToPaint(pos));
      }  
    }
    else if(event->button() == Qt::RightButton) 
    {
      fix(row,col,cellNum);
 //     update();
    }
    else  
        QWidget::mousePressEvent(event);
}
void SudokuWidget::fix(int row, int col, int num,bool allowUndo)
{
  vector<RowCol> toSave;
	if(allowUndo)
		backupSquare(row,col);
	grid[row][col]->fix = num + 1;
	if(allowUndo)
		toSave.push_back(RowCol(row,col));
	update(squareToPaint(row,col));
	
  for(int i=0; i<side; i++) //disable all in same row
  {
	  if(allowUndo)
	  	backupSquare(row,i);
      grid[row][i]->numbers[num] = disabled;
	  if(allowUndo)
	  	toSave.push_back(RowCol(row,i));
	  update(squareToPaint(row,i));
  }
  for(int i=0; i<side; i++) //disable all in same col
  {
	  if(allowUndo)
	  	backupSquare(i,col);
      grid[i][col]->numbers[num] = disabled;
	  if(allowUndo)
	  	toSave.push_back(RowCol(i,col));
	  update(squareToPaint(i,col));
  }
    
  //now disable all in block
  int blockRow = row/squareRootOfSide;
  int blockRowStart = blockRow*squareRootOfSide;
  int blockRowEnd = blockRowStart + squareRootOfSide;
  int blockCol = col/squareRootOfSide;
  int blockColStart = blockCol*squareRootOfSide;
  int blockColEnd = blockColStart + squareRootOfSide;
  
  for(int i=blockRowStart; i< blockRowEnd; i++)
    for(int j=blockColStart; j< blockColEnd; j++)
	{
		if(allowUndo)
			backupSquare(i,j);
        grid[i][j]->numbers[num] = disabled;
		if(allowUndo)
			toSave.push_back(RowCol(i,j));
		update(squareToPaint(i,j));
	}
	if(allowUndo)
		saveChangedSquares(toSave);
}  
void SudokuWidget::resizeEvent ( QResizeEvent * event ) 
{
  calcFontSize();
}  
void SudokuWidget::calcFontSize()
{
   squareWidth = (width())/side;
   float squareWidthForFont = (width()-side*2)/side;
   maxWidth = squareWidth*side;
	cellWidth = squareWidth/squareRootOfSide;
   float cellWidthForFont = (float)(squareWidth- squareRootOfSide*2)/(float)squareRootOfSide;
   //cellWidth = (int) haveWidth;
   
   squareHeight = (height())/side;
   float squareHeightForFont = (height()-side*2)/side;
   maxHeight = squareHeight*side;
	cellHeight = squareHeight/squareRootOfSide;
   float cellHeightForFont = (float)(squareHeight - squareRootOfSide*2)/(float)squareRootOfSide;
   //cellHeight = (int) haveHeight;
   
	 fontSize = fontSizeForRect(cellWidthForFont,cellHeightForFont);
	 bigFontSize = fontSizeForRect(squareWidthForFont,squareHeightForFont);
}  
float SudokuWidget::fontSizeForRect(float haveWidth, float haveHeight)
{
    float initsize = 5.0f;
    displayFont.setPointSizeF(initsize);
		QFontMetrics fm(displayFont);
		QString biggestNum = QString::number(side);
		QRect rect = fm.boundingRect(biggestNum);

		if((float)rect.width()/(float)rect.height() > (float)haveWidth/(float)haveHeight)
		{
			float scale = (float)haveWidth/(float)rect.width();
			return scale*initsize;
		}
		else
		{
			float scale = (float)haveHeight/(float)rect.height();
			return scale*initsize;
		}
}  
void SudokuWidget::paintOnDevice(QPaintDevice* paintDevice,QRect rect)
{
	 QPainter painter(paintDevice);
    
    displayFont.setPointSizeF(fontSize);
    painter.setFont(displayFont);

    painter.setPen(QPen(Qt::gray));
    for(int i=0; i<= side; i++)
    {        
      painter.drawLine(1,i*squareHeight,maxWidth,i*squareHeight); //row
      painter.drawLine(i*squareWidth,1,i*squareWidth,maxHeight); //col
    }  
    
    painter.setPen(QPen(Qt::black));
    for(int i=0; i<= side; i+=squareRootOfSide)
    {
      painter.drawLine(1,i*squareHeight,maxWidth,i*squareHeight); //row
      painter.drawLine(i*squareWidth,1,i*squareWidth,maxHeight); //col
      
    }  
    
    //now figure out which blocks need redrawing
    
    //~ int fromSquareRow = event->rect().top()/squareHeight;
    //~ int toSquareRow = event->rect().bottom()/squareHeight;
    //~ if(toSquareRow >= side)
        //~ toSquareRow = side-1;
    
    //~ int fromSquareCol = event->rect().left()/squareWidth;
    //~ int toSquareCol = event->rect().right()/squareWidth;
    //~ if(toSquareCol >= side)
        //~ toSquareCol = side-1;
	
    int fromSquareRow = rect.top()/squareHeight;
    int toSquareRow = rect.bottom()/squareHeight;
    if(toSquareRow >= side)
        toSquareRow = side-1;
    
    int fromSquareCol = rect.left()/squareWidth;
    int toSquareCol = rect.right()/squareWidth;
    if(toSquareCol >= side)
        toSquareCol = side-1;

    
    for(int i=fromSquareRow; i < toSquareRow+1; i++) 
      for(int j=fromSquareCol; j < toSquareCol+1; j++)
      {
        int squareX = j*squareWidth + 1 ;//+ cellWidth/4; //it seems to draw quite to the left of the box
        int squareY = i*squareHeight + 1 ;//+ cellHeight;
        if(grid[i][j]->fix > 0)
        {
          displayFont.setPointSizeF(bigFontSize);
          painter.setFont(displayFont);
          if(grid[i][j]->reallyFixed)
            painter.setPen(QPen(Qt::green));
          else
            painter.setPen(QPen(Qt::black));
            
          painter.drawText(squareX,squareY,squareWidth,squareHeight,Qt::AlignCenter,numbersAsStrings[grid[i][j]->fix-1]);   
          
          displayFont.setPointSizeF(fontSize);
          painter.setFont(displayFont);
        }
        else  
        {
          for(int p=0; p < squareRootOfSide; p++)
              for(int q=0; q < squareRootOfSide; q++)
              {
                int cellX = squareX + (cellWidth)*q +1;
                int cellY = squareY + (cellHeight)*p + 1;
                if(highlight->active &&
                              highlight->row == i &&
                              highlight->col == j &&
                              highlight->cellRow == p &&
                              highlight->cellCol == q)
                {
                  if(grid[i][j]->numbers[p*squareRootOfSide+q] == on)
                    painter.setPen(QPen(Qt::darkBlue));
                  else
                    painter.setPen(QPen(Qt::blue));
                }
                else
                {
                  if(grid[i][j]->numbers[p*squareRootOfSide+q] == on)
                    painter.setPen(QPen(Qt::black));
                  else if(grid[i][j]->numbers[p*squareRootOfSide+q] == off)
                     painter.setPen(QPen(Qt::lightGray));
                  else
                     painter.setPen(QPen(Qt::white));
                } 
                painter.drawText(cellX,cellY,cellWidth,cellHeight,Qt::AlignCenter,numbersAsStrings[squareRootOfSide*p + q]);   
                
              } 
          }   
      }  
 
}
void SudokuWidget::paintEvent(QPaintEvent *event)
{
	//QRect rect = *(event->rect());
	paintOnDevice(this,event->rect());
       
}
